@@ -1,7 +1,7 @@
 //
 // PDF file functions for PDFio.
 //
-// Copyright © 2021-2025 by Michael R Sweet.
+// Copyright © 2021-2026 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
@@ -2006,6 +2006,9 @@ load_xref(
   _pdfio_token_t tb;			// Token buffer/stack
   off_t		line_offset;		// Offset to start of line
   pdfio_obj_t	*pages_obj;		// Pages object
+  size_t	num_xrefs = 1;		// Number of xref offsets
+  off_t		xrefs[100] = { xref_offset };
+					// xref offsets
 
 
   while (!done)
@@ -2223,20 +2226,26 @@ load_xref(
 	  // Create a placeholder for the object in memory...
 	  if ((current = pdfioFileFindObj(pdf, (size_t)number)) != NULL)
 	  {
-	    PDFIO_DEBUG("load_xref: existing object, prev offset=%u\n", (unsigned)current->offset);
+	    PDFIO_DEBUG("load_xref: existing object, prev offset=%u, generation=%u, new generation=%u\n", (unsigned)current->offset, (unsigned)current->generation, (unsigned)generation);
 
-            if (w[0] == 0 || buffer[0] == 1)
+            if (generation > current->generation)
             {
-              // Location of object...
-	      current->offset = (off_t)offset;
-	    }
-	    else if (number != offset)
-	    {
-	      // Object is part of a stream, offset is the object number...
-	      current->offset = 0;
-	    }
+              // Newer version of an existing object - update the references...
+              current->generation = generation;
 
-	    PDFIO_DEBUG("load_xref: new offset=%u\n", (unsigned)current->offset);
+	      if (w[0] == 0 || buffer[0] == 1)
+	      {
+		// Location of object...
+		current->offset = (off_t)offset;
+	      }
+	      else if (number != offset)
+	      {
+		// Object is part of a stream, offset is the object number...
+		current->offset = 0;
+	      }
+
+	      PDFIO_DEBUG("load_xref: new offset=%u\n", (unsigned)current->offset);
+	    }
 	  }
 
 	  if (w[0] > 0 && buffer[0] == 2)
@@ -2471,13 +2480,31 @@ load_xref(
     {
       done = true;
     }
-    else if (new_offset == xref_offset)
+    else
     {
-      _pdfioFileError(pdf, "Recursive xref table.");
-      return (false);
-    }
+      // See if we've seen this xref table before...
+      size_t	i;			// Looping var
 
-    xref_offset = new_offset;
+      for (i = 0; i < num_xrefs; i ++)
+      {
+        if (new_offset == xrefs[i])
+	{
+	  // Yes, error out...
+	  _pdfioFileError(pdf, "Recursive xref table.");
+	  return (false);
+	}
+      }
+
+      // No, save it...
+      if (i >= (sizeof(xrefs) / sizeof(xrefs[0])))
+      {
+        // Too many xref tables...
+	_pdfioFileError(pdf, "Too many xref tables.");
+	return (false);
+      }
+
+      xrefs[num_xrefs ++] = xref_offset = new_offset;
+    }
   }
 
   // Once we have all of the xref tables loaded, get the important objects and
@@ -2600,19 +2627,19 @@ repair_xref(
 
 	    _pdfioTokenFlush(&tb);
 
-            if (type && !strcmp(line, "stream"))
+            if (!strcmp(line, "stream"))
             {
               // Possible object or XRef stream...
 	      obj->stream_offset = _pdfioFileTell(pdf);
 
-	      if (!strcmp(type, "ObjStm") && num_sobjs < (sizeof(sobjs) / sizeof(sobjs[0])))
+	      if (type && !strcmp(type, "ObjStm") && num_sobjs < (sizeof(sobjs) / sizeof(sobjs[0])))
 	      {
 	        PDFIO_DEBUG("repair_xref: Object stream...\n");
 		sobjs[num_sobjs] = obj;
 		num_sobjs ++;
 	      }
 
-	      if (!strcmp(type, "XRef") && !pdf->trailer_dict)
+	      if (type && !strcmp(type, "XRef") && !pdf->trailer_dict)
 	      {
 		// Save the trailer dictionary...
 		pdfio_obj_t *encrypt_obj;
