@@ -1,7 +1,7 @@
 //
 // PDF value functions for PDFio.
 //
-// Copyright © 2021-2025 by Michael R Sweet.
+// Copyright © 2021-2026 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
@@ -46,7 +46,7 @@ _pdfioValueCopy(pdfio_file_t   *pdfdst,	// I - Destination PDF file
 #endif // DEBUG
 
 
-  PDFIO_DEBUG("_pdfioValueCopy(pdfdst=%p, vdst=%p, pdfsrc=%p, vsrc=%p(%s))\n", pdfdst, vdst, pdfsrc, vsrc, types[vsrc->type]);
+  PDFIO_DEBUG("_pdfioValueCopy(pdfdst=%p, vdst=%p, pdfsrc=%p, vsrc=%p(%s))\n", (void *)pdfdst, (void *)vdst, (void *)pdfsrc, (void *)vsrc, types[vsrc->type]);
 
   if (pdfdst == pdfsrc && vsrc->type != PDFIO_VALTYPE_BINARY)
   {
@@ -76,7 +76,8 @@ _pdfioValueCopy(pdfio_file_t   *pdfdst,	// I - Destination PDF file
         return (NULL);
 
     case PDFIO_VALTYPE_ARRAY :
-        vdst->value.array = pdfioArrayCopy(pdfdst, vsrc->value.array);
+        if ((vdst->value.array = pdfioArrayCopy(pdfdst, vsrc->value.array)) == NULL)
+          return (NULL);
         break;
 
     case PDFIO_VALTYPE_BINARY :
@@ -97,12 +98,14 @@ _pdfioValueCopy(pdfio_file_t   *pdfdst,	// I - Destination PDF file
         return (vdst);
 
     case PDFIO_VALTYPE_DICT :
-        vdst->value.dict = pdfioDictCopy(pdfdst, vsrc->value.dict);
+        if ((vdst->value.dict = pdfioDictCopy(pdfdst, vsrc->value.dict)) == NULL)
+          return (NULL);
         break;
 
     case PDFIO_VALTYPE_NAME :
     case PDFIO_VALTYPE_STRING :
-        vdst->value.name = pdfioStringCreate(pdfdst, vsrc->value.name);
+        if ((vdst->value.name = pdfioStringCreate(pdfdst, vsrc->value.name)) == NULL)
+          return (NULL);
         break;
   }
 
@@ -125,7 +128,7 @@ _pdfioValueDecrypt(pdfio_file_t   *pdf,	// I - PDF file
   _pdfio_crypto_ctx_t	ctx;		// Decryption context
   _pdfio_crypto_cb_t	cb;		// Decryption callback
   size_t		ivlen;		// Number of initialization vector bytes
-  uint8_t		temp[32768];	// Temporary buffer for decryption
+  uint8_t		*temp = NULL;	// Temporary buffer for decryption
   size_t		templen;	// Number of actual data bytes
   time_t		timeval;	// Date/time value
 
@@ -152,11 +155,16 @@ _pdfioValueDecrypt(pdfio_file_t   *pdf,	// I - PDF file
 
     case PDFIO_VALTYPE_BINARY :
 	// Decrypt the binary string...
-	if (v->value.binary.datalen > (sizeof(temp) - 32))
+	if (v->value.binary.datalen > PDFIO_MAX_STRING)
 	{
 	  _pdfioFileError(pdf, "Unable to read encrypted binary string - too long.");
 	  return (false);
 	}
+	else if ((temp = (uint8_t *)_pdfioStringAllocBuffer(pdf)) == NULL)
+	{
+	  _pdfioFileError(pdf, "Unable to read encrypted binary string - out of memory.");
+	  return (false);
+	}
 
 	ivlen = v->value.binary.datalen;
 	if ((cb = _pdfioCryptoMakeReader(pdf, obj, &ctx, v->value.binary.data, &ivlen)) == NULL)
@@ -167,29 +175,59 @@ _pdfioValueDecrypt(pdfio_file_t   *pdf,	// I - PDF file
 	// Copy the decrypted string back to the value and adjust the length...
 	memcpy(v->value.binary.data, temp, templen);
 
-	if (pdf->encryption >= PDFIO_ENCRYPTION_AES_128)
+	if (pdf->encryption >= PDFIO_ENCRYPTION_AES_128 && temp[templen - 1] <= templen)
 	  v->value.binary.datalen = templen - temp[templen - 1];
 	else
 	  v->value.binary.datalen = templen;
+
+        _pdfioStringFreeBuffer(pdf, (char *)temp);
 	break;
 
     case PDFIO_VALTYPE_STRING :
         // Decrypt regular string...
         templen = strlen(v->value.string);
-	if (templen > (sizeof(temp) - 33))
+	if (templen > (PDFIO_MAX_STRING - 1))
 	{
 	  _pdfioFileError(pdf, "Unable to read encrypted string - too long.");
 	  return (false);
 	}
+	else if ((temp = (uint8_t *)_pdfioStringAllocBuffer(pdf)) == NULL)
+	{
+	  _pdfioFileError(pdf, "Unable to read encrypted binary string - out of memory.");
+	  return (false);
+	}
 
 	ivlen = templen;
 	if ((cb = _pdfioCryptoMakeReader(pdf, obj, &ctx, (uint8_t *)v->value.string, &ivlen)) == NULL)
 	  return (false);
 
 	templen = (cb)(&ctx, temp, (uint8_t *)v->value.string + ivlen, templen - ivlen);
+
+	if (pdf->encryption >= PDFIO_ENCRYPTION_AES_128 && temp[templen - 1] <= templen)
+	  templen -= temp[templen - 1];
+
 	temp[templen] = '\0';
 
-        if ((timeval = get_date_time((char *)temp)) != 0)
+        if ((templen & 1) == 0 && (!memcmp(temp, "\376\377", 2) || !memcmp(temp, "\377\376", 2)))
+        {
+          // Convert UTF-16 to UTF-8...
+          char	utf8[4096];		// Temporary string
+
+          _pdfio_utf16cpy(utf8, temp, templen, sizeof(utf8));
+
+	  if ((timeval = get_date_time((char *)utf8)) != 0)
+	  {
+	    // Change the type to date...
+	    v->type       = PDFIO_VALTYPE_DATE;
+	    v->value.date = timeval;
+	  }
+	  else
+	  {
+	    // Copy the decrypted string back to the value...
+	    v->value.string = pdfioStringCreate(pdf, utf8);
+	  }
+        }
+        else if ((timeval = get_date_time((char *)temp)) != 0)
         {
           // Change the type to date...
           v->type       = PDFIO_VALTYPE_DATE;
@@ -200,6 +238,8 @@ _pdfioValueDecrypt(pdfio_file_t   *pdf,	// I - PDF file
           // Copy the decrypted string back to the value...
 	  v->value.string = pdfioStringCreate(pdf, (char *)temp);
 	}
+
+        _pdfioStringFreeBuffer(pdf, (char *)temp);
         break;
   }
 
@@ -300,7 +340,9 @@ _pdfioValueRead(pdfio_file_t   *pdf,	// I - PDF file
                 _pdfio_value_t *v,	// I - Value
                 size_t         depth)	// I - Depth of value
 {
-  char		token[32768];		// Token buffer
+  _pdfio_value_t *ret = NULL;		// Return value
+  char		*token = _pdfioStringAllocBuffer(pdf);
+					// Token buffer
   time_t	timeval;		// Date/time value
 #ifdef DEBUG
   static const char * const valtypes[] =
@@ -320,10 +362,13 @@ _pdfioValueRead(pdfio_file_t   *pdf,	// I - PDF file
 #endif // DEBUG
 
 
-  PDFIO_DEBUG("_pdfioValueRead(pdf=%p, obj=%p, v=%p)\n", pdf, obj, v);
+  PDFIO_DEBUG("_pdfioValueRead(pdf=%p, obj=%p, v=%p)\n", (void *)pdf, (void *)obj, (void *)v);
 
-  if (!_pdfioTokenGet(tb, token, sizeof(token)))
-    return (NULL);
+  if (!token)
+    goto done;
+
+  if (!_pdfioTokenGet(tb, token, PDFIO_MAX_STRING))
+    goto done;
 
   if (!strcmp(token, "["))
   {
@@ -331,12 +376,14 @@ _pdfioValueRead(pdfio_file_t   *pdf,	// I - PDF file
     if (depth >= PDFIO_MAX_DEPTH)
     {
       _pdfioFileError(pdf, "Too many nested arrays.");
-      return (NULL);
+      goto done;
     }
 
     v->type = PDFIO_VALTYPE_ARRAY;
     if ((v->value.array = _pdfioArrayRead(pdf, obj, tb, depth + 1)) == NULL)
-      return (NULL);
+      goto done;
+
+    ret = v;
   }
   else if (!strcmp(token, "<<"))
   {
@@ -344,29 +391,38 @@ _pdfioValueRead(pdfio_file_t   *pdf,	// I - PDF file
     if (depth >= PDFIO_MAX_DEPTH)
     {
       _pdfioFileError(pdf, "Too many nested dictionaries.");
-      return (NULL);
+      goto done;
     }
 
     v->type = PDFIO_VALTYPE_DICT;
     if ((v->value.dict = _pdfioDictRead(pdf, obj, tb, depth + 1)) == NULL)
-      return (NULL);
-  }
-  else if (!strncmp(token, "(D:", 3) && (timeval = get_date_time(token + 1)) != 0)
-  {
-    v->type       = PDFIO_VALTYPE_DATE;
-    v->value.date = timeval;
+      goto done;
+
+    ret = v;
   }
   else if (token[0] == '(')
   {
-    // String
-    v->type         = PDFIO_VALTYPE_STRING;
-    v->value.string = pdfioStringCreate(pdf, token + 1);
+    if ((timeval = get_date_time(token + 1)) != 0)
+    {
+      // Date
+      v->type       = PDFIO_VALTYPE_DATE;
+      v->value.date = timeval;
+      ret           = v;
+    }
+    else
+    {
+      // String
+      v->type         = PDFIO_VALTYPE_STRING;
+      v->value.string = pdfioStringCreate(pdf, token + 1);
+      ret           = v;
+    }
   }
   else if (token[0] == '/')
   {
     // Name
     v->type       = PDFIO_VALTYPE_NAME;
     v->value.name = pdfioStringCreate(pdf, token + 1);
+    ret           = v;
   }
   else if (token[0] == '<')
   {
@@ -379,7 +435,7 @@ _pdfioValueRead(pdfio_file_t   *pdf,	// I - PDF file
     if ((v->value.binary.data = (unsigned char *)malloc(v->value.binary.datalen)) == NULL)
     {
       _pdfioFileError(pdf, "Out of memory for hex string.");
-      return (NULL);
+      goto done;
     }
 
     // Convert hex to binary...
@@ -406,6 +462,8 @@ _pdfioValueRead(pdfio_file_t   *pdf,	// I - PDF file
 
       *dataptr++ = (unsigned char)d;
     }
+
+    ret = v;
   }
   else if (strchr("0123456789-+.", token[0]) != NULL)
   {
@@ -460,7 +518,7 @@ _pdfioValueRead(pdfio_file_t   *pdf,	// I - PDF file
         // Integer...
         long generation = 0;		// Generation number
 
-        while (tempptr < tb->bufend && isdigit(*tempptr & 255))
+        while (tempptr < tb->bufend && generation < 65536 && isdigit(*tempptr & 255))
         {
           generation = generation * 10 + *tempptr - '0';
           tempptr ++;
@@ -493,7 +551,8 @@ _pdfioValueRead(pdfio_file_t   *pdf,	// I - PDF file
 
 	  PDFIO_DEBUG("_pdfioValueRead: Returning indirect value %lu %u R.\n", (unsigned long)v->value.indirect.number, v->value.indirect.generation);
 
-	  return (v);
+	  ret = v;
+	  goto done;
 	}
       }
     }
@@ -501,27 +560,41 @@ _pdfioValueRead(pdfio_file_t   *pdf,	// I - PDF file
     // If we get here, we have a number...
     v->type         = PDFIO_VALTYPE_NUMBER;
     v->value.number = _pdfio_strtod(pdf, token);
+    ret             = v;
   }
   else if (!strcmp(token, "true") || !strcmp(token, "false"))
   {
     // Boolean value
     v->type          = PDFIO_VALTYPE_BOOLEAN;
     v->value.boolean = !strcmp(token, "true");
+    ret              = v;
   }
   else if (!strcmp(token, "null"))
   {
     // null value
     v->type = PDFIO_VALTYPE_NULL;
+    ret     = v;
   }
   else
   {
     _pdfioFileError(pdf, "Unexpected '%s' token seen.", token);
-    return (NULL);
   }
 
-  PDFIO_DEBUG("_pdfioValueRead: Returning %s value.\n", valtypes[v->type]);
+  done:
 
-  return (v);
+  if (token)
+    _pdfioStringFreeBuffer(pdf, token);
+
+  if (ret)
+  {
+    PDFIO_DEBUG("_pdfioValueRead: Returning %s value.\n", valtypes[ret->type]);
+    return (ret);
+  }
+  else
+  {
+    PDFIO_DEBUG("_pdfioValueRead: Returning NULL.\n");
+    return (NULL);
+  }
 }
 
 
@@ -546,8 +619,10 @@ _pdfioValueWrite(pdfio_file_t   *pdf,	// I - PDF file
     case PDFIO_VALTYPE_BINARY :
         {
           size_t	databytes;	// Bytes to write
-          uint8_t	temp[32768],	// Temporary buffer for encryption
+          uint8_t	*temp = NULL,	// Temporary buffer for encryption
 			*dataptr;	// Pointer into data
+          bool          ret = false;	// Return value
+
 
           if (obj && pdf->encryption)
           {
@@ -556,11 +631,16 @@ _pdfioValueWrite(pdfio_file_t   *pdf,	// I - PDF file
 	    _pdfio_crypto_cb_t cb;	// Encryption callback
 	    size_t	ivlen;		// Number of initialization vector bytes
 
-            if (v->value.binary.datalen > (sizeof(temp) - 32))
+            if (v->value.binary.datalen > PDFIO_MAX_STRING)
             {
 	      _pdfioFileError(pdf, "Unable to write encrypted binary string - too long.");
 	      return (false);
             }
+            else if ((temp = (uint8_t *)_pdfioStringAllocBuffer(pdf)) == NULL)
+            {
+	      _pdfioFileError(pdf, "Unable to write encrypted binary string - out of memory.");
+	      return (false);
+            }
 
 	    cb        = _pdfioCryptoMakeWriter(pdf, obj, &ctx, temp, &ivlen);
 	    databytes = (cb)(&ctx, temp + ivlen, v->value.binary.data, v->value.binary.datalen) + ivlen;
@@ -573,18 +653,25 @@ _pdfioValueWrite(pdfio_file_t   *pdf,	// I - PDF file
           }
 
           if (!_pdfioFilePuts(pdf, "<"))
-            return (false);
+            goto bindone;
 
           for (; databytes > 1; databytes -= 2, dataptr += 2)
           {
             if (!_pdfioFilePrintf(pdf, "%02X%02X", dataptr[0], dataptr[1]))
-              return (false);
+              goto bindone;
           }
 
-          if (databytes > 0)
-            return (_pdfioFilePrintf(pdf, "%02X>", dataptr[0]));
-          else
-            return (_pdfioFilePuts(pdf, ">"));
+          if (databytes > 0 && !_pdfioFilePrintf(pdf, "%02X", dataptr[0]))
+            goto bindone;
+
+	  ret = _pdfioFilePuts(pdf, ">");
+
+          bindone:
+
+          if (temp)
+            _pdfioStringFreeBuffer(pdf, (char *)temp);
+
+          return (ret);
         }
 
     case PDFIO_VALTYPE_BOOLEAN :
@@ -609,7 +696,7 @@ _pdfioValueWrite(pdfio_file_t   *pdf,	// I - PDF file
 	  if (obj && pdf->encryption)
 	  {
 	    // Write encrypted string...
-	    uint8_t	temp[32768],	// Encrypted bytes
+	    uint8_t	temp[64],	// Encrypted bytes
 			*tempptr;	// Pointer into encrypted bytes
 	    _pdfio_crypto_ctx_t ctx;	// Encryption context
 	    _pdfio_crypto_cb_t cb;	// Encryption callback
@@ -660,7 +747,7 @@ _pdfioValueWrite(pdfio_file_t   *pdf,	// I - PDF file
         if (obj && pdf->encryption)
         {
           // Write encrypted string...
-          uint8_t	temp[32768],	// Encrypted bytes
+          uint8_t	*temp = NULL,	// Encrypted bytes
 			*tempptr;	// Pointer into encrypted bytes
           _pdfio_crypto_ctx_t ctx;	// Encryption context
           _pdfio_crypto_cb_t cb;	// Encryption callback
@@ -668,29 +755,41 @@ _pdfioValueWrite(pdfio_file_t   *pdf,	// I - PDF file
 					// Length of value
 			ivlen,		// Number of initialization vector bytes
 			tempbytes;	// Number of output bytes
+	  bool		ret = false;	// Return value
 
-          if (len > (sizeof(temp) - 32))
+          if (len > PDFIO_MAX_STRING)
           {
             _pdfioFileError(pdf, "Unable to write encrypted string - too long.");
             return (false);
           }
+          else if ((temp = (uint8_t *)_pdfioStringAllocBuffer(pdf)) == NULL)
+          {
+            _pdfioFileError(pdf, "Unable to write encrypted string - out of memory.");
+            return (false);
+          }
 
           cb        = _pdfioCryptoMakeWriter(pdf, obj, &ctx, temp, &ivlen);
           tempbytes = (cb)(&ctx, temp + ivlen, (const uint8_t *)v->value.string, len) + ivlen;
 
           if (!_pdfioFilePuts(pdf, "<"))
-            return (false);
+            goto strdone;
 
           for (tempptr = temp; tempbytes > 1; tempbytes -= 2, tempptr += 2)
           {
             if (!_pdfioFilePrintf(pdf, "%02X%02X", tempptr[0], tempptr[1]))
-              return (false);
+              goto strdone;
           }
 
-          if (tempbytes > 0)
-            return (_pdfioFilePrintf(pdf, "%02X>", *tempptr));
-          else
-	    return (_pdfioFilePuts(pdf, ">"));
+          if (tempbytes > 0 && !_pdfioFilePrintf(pdf, "%02X", *tempptr))
+            goto strdone;
+
+	  ret = _pdfioFilePuts(pdf, ">");
+
+          strdone :
+
+          _pdfioStringFreeBuffer(pdf, (char *)temp);
+
+          return (ret);
         }
         else
         {
@@ -707,31 +806,59 @@ _pdfioValueWrite(pdfio_file_t   *pdf,	// I - PDF file
 // 'get_date_time()' - Convert PDF date/time value to time_t.
 //
 
-static time_t				// O - Time in seconds
+static time_t				// O - Time in seconds or `0` for none
 get_date_time(const char *s)		// I - PDF date/time value
 {
   int		i;			// Looping var
   struct tm	dateval;		// Date value
-  int		offset;			// Date offset
+  int		offset = 0;		// Date offset in seconds
+  time_t	t;			// Time value
 
 
   PDFIO_DEBUG("get_date_time(s=\"%s\")\n", s);
 
   // Possible date value of the form:
   //
-  //   (D:YYYYMMDDhhmmssZ)
-  //   (D:YYYYMMDDhhmmss+HH'mm)
-  //   (D:YYYYMMDDhhmmss-HH'mm)
+  //   D:YYYYMMDDhhmmssZ
+  //   D:YYYYMMDDhhmmss+HH'mm
+  //   D:YYYYMMDDhhmmss-HH'mm
   //
 
+  if (strncmp(s, "D:", 2))
+    return (0);
+
   for (i = 2; i < 16; i ++)
   {
+    // Look for date/time digits...
     if (!isdigit(s[i] & 255) || !s[i])
       break;
   }
 
-  if (i >= 16)
+  if (i < 6 || (i & 1))
   {
+    // Short year or missing digit...
+    return (0);
+  }
+
+  memset(&dateval, 0, sizeof(dateval));
+
+  dateval.tm_year = (s[2] - '0') * 1000 + (s[3] - '0') * 100 + (s[4] - '0') * 10 + s[5] - '0' - 1900;
+  if (i > 6)
+    dateval.tm_mon  = (s[6] - '0') * 10 + s[7] - '0' - 1;
+  if (i > 8)
+    dateval.tm_mday = (s[8] - '0') * 10 + s[9] - '0';
+  else
+    dateval.tm_mday = 1;
+  if (i > 10)
+    dateval.tm_hour = (s[10] - '0') * 10 + s[11] - '0';
+  if (i > 12)
+    dateval.tm_min  = (s[12] - '0') * 10 + s[13] - '0';
+  if (i > 14)
+    dateval.tm_sec  = (s[14] - '0') * 10 + s[15] - '0';
+
+  if (i >= 16 && s[i])
+  {
+    // Get zone info...
     if (s[i] == 'Z')
     {
       // UTC...
@@ -742,14 +869,20 @@ get_date_time(const char *s)		// I - PDF date/time value
       // Timezone offset from UTC...
       if (isdigit(s[i + 1] & 255) && isdigit(s[i + 2] & 255) && s[i + 3] == '\'' && isdigit(s[i + 4] & 255) && isdigit(s[i + 5] & 255))
       {
+	offset = (s[i + 1] - '0') * 36000 + (s[i + 2] - '0') * 3600 + (s[i + 4] - '0') * 600 + (s[i + 5] - '0') * 60;
+	if (s[i] == '-')
+	  offset = -offset;
+
 	i += 6;
+
+	// Accept trailing quote, per PDF spec...
 	if (s[i] == '\'')
 	  i ++;
       }
     }
-    else if (!s[i])
+    else
     {
-      // Missing zone info, invalid date string...
+      // Random zone info, invalid date string...
       return (0);
     }
   }
@@ -760,26 +893,31 @@ get_date_time(const char *s)		// I - PDF date/time value
     return (0);
   }
 
-  // Date value...
-  memset(&dateval, 0, sizeof(dateval));
+  // Convert date value to time_t...
+#if _WIN32
+  if ((t = _mkgmtime(&dateval)) <= 0)
+    return (0);
 
-  dateval.tm_year = (s[2] - '0') * 1000 + (s[3] - '0') * 100 + (s[4] - '0') * 10 + s[5] - '0' - 1900;
-  dateval.tm_mon  = (s[6] - '0') * 10 + s[7] - '0' - 1;
-  dateval.tm_mday = (s[8] - '0') * 10 + s[9] - '0';
-  dateval.tm_hour = (s[10] - '0') * 10 + s[11] - '0';
-  dateval.tm_min  = (s[12] - '0') * 10 + s[13] - '0';
-  dateval.tm_sec  = (s[14] - '0') * 10 + s[15] - '0';
+#elif defined(HAVE_TIMEGM)
+  if ((t = timegm(&dateval)) <= 0)
+    return (0);
 
-  if (s[16] == 'Z')
-  {
-    offset = 0;
-  }
-  else
-  {
-    offset = (s[17] - '0') * 600 + (s[18] - '0') * 60 + (s[19] - '0') * 10 + s[20] - '0';
-    if (s[16] == '-')
-      offset = -offset;
-  }
+#else
+  if ((t = mktime(&dateval)) <= 0)
+    return (0);
 
-  return (mktime(&dateval) + offset);
+#  if defined(HAVE_TM_GMTOFF)
+  // Adjust the time value using the "tm_gmtoff" and "tm_isdst" members.  As
+  // noted by M-HT on Github, this DST hack will fail in timezones where the
+  // DST offset is not one hour, such as Australia/Lord_Howe.  Fortunately,
+  // this is unusual and most systems support the "timegm" function...
+  t += dateval.tm_gmtoff - 3600 * dateval.tm_isdst;
+#  else
+  // Adjust the time value using the even more legacy "timezone" variable,
+  // which also reflects any DST offset...
+  t += timezone;
+#  endif // HAVE_TM_GMTOFF
+#endif // _WIN32
+
+  return (t - offset);
 }
